;;; -*- Mode: Lisp; indent-tabs-mode: nil -*-
;;    cl-arcball: an arcball implementation in Common Lisp.
;;    Best used with OpenGL and GLUT or FreeGlut.
;;
;;    Copyright (C) 2006  Sathyaraju Fuentes Gnanasekaran
;;
;;    This library is free software; you can redistribute it and/or
;;    modify it under the terms of the GNU Lesser General Public
;;    License as published by the Free Software Foundation; either
;;    version 2.1 of the License, or (at your option) any later version.
;;
;;    This library is distributed in the hope that it will be useful,
;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;    Lesser General Public License for more details.
;;
;;    You should have received a copy of the GNU Lesser General Public
;;    License along with this library; if not, write to the Free Software
;;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

;;; package
(defpackage #:cl-arcball
  (:use #:common-lisp)
  (:nicknames #:arcball)
  (:export
   #:multquats
   #:quat-to-matrix
   #:crossprod
   #:dotprod
   #:arcball
   #:matrix
   #:clickvec
   #:width
   #:height
   #:rotation
   #:lastrotation
   #:make-arcball
   #:maptosphere
   #:click
   #:drag
   #:setbounds
   #:reset
   #:rotate))

(in-package #:cl-arcball)

;;;; functions for quaternions

(defun multquats (left right)
  `(,(- (* (first left) (first right))
	(* (second left) (second right))
	(* (third left) (third right))
	(* (fourth left) (fourth right)))
     ,(+ (* (first left) (second right))
	 (* (second left) (first right))
	 (* (third left) (fourth right))
	 (- (* (fourth left) (third right))))
     ,(+ (* (first left) (third right))
	 (* (third left) (first right))
	 (* (fourth left) (second right))
	 (- (* (second left) (fourth right))))
     ,(+ (* (first left) (fourth right))
	 (* (fourth left) (first right))
	 (* (second left) (third right))
	 (- (* (third left) (second right))))))
		
(defun quat-to-matrix (quat)
  (let 
      ((xx (expt (second quat) 2))
       (yy (expt (third quat) 2))
       (zz (expt (fourth quat) 2))
       (xy (* (second quat) (third quat)))
       (xz (* (second quat) (fourth quat)))
       (yz (* (third quat) (fourth quat)))
       (wx (* (first quat) (second quat)))
       (wy (* (first quat) (third quat)))
       (wz (* (first quat) (fourth quat))))
    (make-array
     '(4 4)
     :initial-contents
     `((,(- 1.0 (* 2.0 (+ yy zz))) ,(* 2.0 (- xy wz)) ,(* 2.0 (+ xz wy)) 0.0)
       (,(* 2.0 (+ xy wz)) ,(- 1.0 (* 2.0 (+ xx zz))) ,(* 2.0 (- yz wx)) 0.0)
       (,(* 2.0 (- xz wy)) ,(* 2.0 (+ yz wx)) ,(- 1.0 (* 2.0 (+ xx yy))) 0.0)
       (0.0 0.0 0.0 1.0)))))


;;;; functions for R3 vectors

; crossprod only works for R^3 vectors
(defun crossprod (left right)
  `(,(- (* (second left) (third right)) (* (third left) (second right))) ,(- (* (third left) (first right)) (* (first left) (third right))) ,(- (* (first left) (second right)) (* (second left) (first right)))))

; dotprod works for all R^n vectors, but we'll only use it for R3 here
(defun dotprod (left right)
  (apply #'+ (mapcar #'* left right)))

;;;; arcball class/ methods

(defclass arcball ()
  ((matrix :accessor matrix :initform (make-array '(4 4) :initial-contents '((1.0 0.0 0.0 0.0) (0.0 1.0 0.0 0.0) (0.0 0.0 1.0 0.0) (0.0 0.0 0.0 1.0))) :initarg :matrix)
   (clickvec :accessor clickvec :initform '(0 0 0) :initarg :clickvec)
   (width :accessor width :initform 0 :initarg :width)
   (height :accessor height :initform 0 :initarg :height)
   (rotation :accessor rotation :initform '(1 0 0 0) :initarg :rotation)
   (lastrotation :accessor lastrotation :initform '(1 0 0 0) :initarg :lastrotation)))

;basic instance-creator / initializer
(defun make-arcball ()
  (make-instance 'arcball))

;maps a 2d vector to a 3d vector on an arcball
(defmethod maptosphere ((ball arcball) (x integer) (y integer))
  "Maps a 2d vector to a 3d vector"
  (let ((length (+ (expt (- 1.0 (* x (width ball))) 2) (expt (- (* y (height ball)) 1.0) 2))))
    (if (> length 1.0)
        (let ((norm (/ 1.0 (sqrt length))))
          `(,(* norm (- 1.0 (* x (width ball)))) ,(* norm (- (* y (height ball)) 1.0)) 0.0))
        (list (- 1.0 (* x (width ball))) (- (* y (height ball)) 1.0) (sqrt (- 1.0 length))))))

;perform some setup when the user clicks
(defmethod click ((ball arcball) (x integer) (y integer))
  (setf (clickvec ball) (maptosphere ball x y))
  (setf (lastrotation ball) (multquats (lastrotation ball) (rotation ball)))
  (setf (rotation ball) '(1 0 0 0))
  t)

;perform an incremental rotation when the user drags
(defmethod drag ((ball arcball) (x integer) (y integer))
  (let 
      ((drag (maptosphere ball x y)))
    (rotate ball (clickvec ball) drag)
    (setf (clickvec ball) drag)))

;basically, set the aspect ratio of the arcball due to the screen
(defmethod setbounds ((ball arcball) (w integer) (h integer))
  (setf (width ball) (/ 1.0 (* (- w 1.0) 0.5)))
  (setf (height ball) (/ 1.0 (* (- h 1.0) 0.5)))
  t)

(defmethod reset ((ball arcball)) 
  (setf (matrix ball) (make-array 
                       '(4 4)
                       :initial-contents
                       '((1.0 0.0 0.0 0.0)
                         (0.0 1.0 0.0 0.0)
                         (0.0 0.0 1.0 0.0)
                         (0.0 0.0 0.0 1.0))))
  (setf (rotation ball) '(1 0 0 0))
  (setf (lastrotation ball) '(1 0 0 0))
  t)

(defmethod rotate ((ball arcball) start end)
  "Rotate the quaternion using two 3d vectors"
  (let*
  ; convert the crossproduct of the start and end vectors to a quaternion 
      ((quat (cons (dotprod start end) (crossprod start end)))
  ; find the norm of the quaternion
       (norm (sqrt (apply #'+ (mapcar #'expt quat '(2 2 2 2))))))
  ; normalize the quaternion
    (setf quat (mapcar #'/ quat `(,norm ,norm ,norm ,norm)))
  ; set the current rotation
    (setf (rotation ball) (multquats quat (rotation ball)))
  ; find the current rotation from the lastrotation and rotation quaternions
    (setf quat (multquats (lastrotation ball) (rotation ball)))
  ; convert the quaternion to a matrix
    (setf (matrix ball) (quat-to-matrix quat))))
